package checkers.interned;

import checkers.source.*;
import checkers.basetype.*;
import checkers.types.*;
import checkers.util.*;
import com.sun.source.tree.*;
import com.sun.source.util.*;

import javax.lang.model.element.*;
import javax.lang.model.type.*;

/**
 * A type-checking visitor for the {@link checkers.quals.Interned} type
 * qualifier that uses the {@link BaseTypeVisitor} implementation.
 * 
 * @see BaseTypeVisitor
 */
public class InternedVisitor extends BaseTypeVisitor<Void, Void> {

    /** The interned annotation. */
    private final AnnotationMirror INTERNED;
    
    /**
     * Creates a new visitor for type-checking {@link checkers.quals.Interned}.
     *
     * @param checker the {@link BaseTypeChecker} to use
     * @param root the root of the input program's AST to check
     */
    public InternedVisitor(BaseTypeChecker checker, CompilationUnitTree root) {
        super(checker, root);
        this.INTERNED = this.annoFactory.fromName("checkers.quals.Interned");
    }

    @Override
    public Void visitBinary(BinaryTree node, Void p) {

        if (!(node.getKind() == Tree.Kind.EQUAL_TO || 
                node.getKind() == Tree.Kind.NOT_EQUAL_TO))
            return super.visitBinary(node, p);

        Tree leftOp = node.getLeftOperand(), rightOp = node.getRightOperand();

        if (leftOp.getKind() == Tree.Kind.NULL_LITERAL ||
                rightOp.getKind() == Tree.Kind.NULL_LITERAL)
            return super.visitBinary(node, p);

        AnnotatedTypeMirror left = factory.getAnnotatedType(leftOp);
        AnnotatedTypeMirror right = factory.getAnnotatedType(rightOp);

        if (!left.hasAnnotation(INTERNED) || !right.hasAnnotation(INTERNED))
            checker.report(Result.failure("not.interned"), node);

        return super.visitBinary(node, p);
    }

    @Override
    public Void visitMethodInvocation(MethodInvocationTree node, Void p) {

        Element elt = InternalUtils.symbol(node);
        if (elt != null && elt.getKind() == ElementKind.METHOD && 
                "equals".equals(elt.getSimpleName().toString())) {
            TreePath path = trees.getPath(root, node);
            TypeMirror returnType = trees.getTypeMirror(path);
            if (returnType.getKind() == TypeKind.BOOLEAN 
                    && node.getArguments().size() == 1 
                    && node.getMethodSelect() instanceof MemberSelectTree) {

                MemberSelectTree ms = (MemberSelectTree)node.getMethodSelect();
                
                AnnotatedTypeMirror recv = factory.getAnnotatedType(ms.getExpression());
                AnnotatedTypeMirror comp = factory.getAnnotatedType(node.getArguments().get(0));

                if (this.checker.getLintOption("dotequals", true)
                        && recv.hasAnnotation(INTERNED)
                        && comp.hasAnnotation(INTERNED))
                    checker.report(Result.warning("unnecessary.equals"), node);
            }
                
        }

        return super.visitMethodInvocation(node, p);
    }
}
