package checkers.nonnull;

import java.util.*;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.util.SimpleElementVisitor6;

import checkers.nonnull.flow.*;
import checkers.quals.*;
import checkers.types.*;
import checkers.util.InternalUtils;

import com.sun.source.tree.*;
import com.sun.source.util.*;

public class NonNullAnnotatedTypeFactory extends AnnotatedTypeFactory {

    private final FlowVisitor flow;
    private final TypePostAnnotator typePost;
    private final TreePreAnnotator treePre;
    private final ElementPreAnnotator elementPre;
    protected final AnnotationMirror NONNULL;
    protected final AnnotationMirror NULLABLE;
    private final SourcePositions srcPos;
    
    public NonNullAnnotatedTypeFactory(ProcessingEnvironment env,
            CompilationUnitTree root) {
        super(env, root);
        
        typePost = new TypePostAnnotator();
        treePre = new TreePreAnnotator();
        elementPre = new ElementPreAnnotator();
        
        srcPos = trees.getSourcePositions();
        
        NONNULL = this.annotations.fromName("checkers.quals.NonNull");
        NULLABLE = this.annotations.fromName("checkers.quals.Nullable");
        flow = new FlowVisitor(root, srcPos, this, NONNULL);
        flow.scan(root, null);
    }

    @Override
    public AnnotatedTypeMirror getAnnotatedType(Tree tree) {
        
        // Flow-sensitive inference
        boolean flowNN = false;
        Element elt = InternalUtils.symbol(tree);
        if (elt != null) {
            long pos = srcPos.getStartPosition(root, tree);
            for (FlowScope fs : flow.getResults()) {
                if (fs.contains(elt, pos)) {
                    flowNN = true;
                    break;
                }
            }
        }
        
        Tree t = skipParens(tree);
        
        List<AnnotationMirror> preAnnotations = treePre.visit(t,
            new LinkedList<AnnotationMirror>());
        
        AnnotatedTypeMirror result = super.getAnnotatedType(t);
        assert result != null : t + " " + t.getKind();
        
        result.addAnnotations(preAnnotations);
        if (flowNN) result.addAnnotation(NONNULL);
        annotateDefaults(tree, result);
        return complete(result);
    }
    
    private Element nearestEnclosing(Tree tree) {
        
        TreePath path = trees.getPath(root, tree);
        if (path == null)
            return null;
        
        do {
            Tree t = path.getLeaf();
            Element elt = InternalUtils.symbol(t);
            if (elt != null)
                return elt;
        } while ((path = path.getParentPath()) != null);
                
        return null;
    }
    
    private void annotateDefaults(Tree tree, AnnotatedTypeMirror type) {
        
        Element nearest = nearestEnclosing(tree);
        
        // FIXME: apply defaults in more places
        
        // Only for some trees.
        switch (tree.getKind()) {
        case VARIABLE:
            applyDefaults(elementFromDeclaration((VariableTree)tree), type);
            break;
        case NEW_CLASS:
            if (nearest != null)
                applyDefaults(nearest, type);
            break;
        default: // Do nothing.
        }
    }
    
    private void applyDefaults(Element elt, AnnotatedTypeMirror type) {
        assert elt != null : "null element";
        Element e = elt;
        while (e != null) {
            Default d = e.getAnnotation(Default.class);
            if (d != null && d.value().equals(NonNull.class.getName())) {
                  new AnnotatedTypeScanner<Void, AnnotationMirror>() {

                    @Override
                    public Void scan(AnnotatedTypeMirror type,
                            AnnotationMirror p) {
                        if (type != null && !type.hasAnnotation(NULLABLE))
                            type.addAnnotation(p);
                        return super.scan(type, p);
                    }

                }.scan(type, NONNULL);
            }
            e = e.getEnclosingElement();
        }
    }
    
    @Override
    public AnnotatedTypeMirror getAnnotatedType(Element elt) {
        
        Tree decl = declarationFromElement(elt);
        if (decl != null)
            return getAnnotatedType(decl);
        
        List<AnnotationMirror> preAnnotations = elementPre.visit(elt,
                new LinkedList<AnnotationMirror>());
        
        AnnotatedTypeMirror result = super.getAnnotatedType(elt);
        assert result != null : elt + " " + elt.getKind();
        
        result.addAnnotations(preAnnotations);
//        annotateDefaults(elt, result);
        return complete(result);
    }

    private AnnotatedTypeMirror complete(AnnotatedTypeMirror type) {
        typePost.visit(type);
        // checkRep
        return type;
    }
    
    protected class TypePostAnnotator extends AnnotatedTypeScanner<Void, Void> {
        // Empty for now.
    }
    
    protected class TreePreAnnotator extends SimpleTreeVisitor<List<AnnotationMirror>, List<AnnotationMirror>> {
        
        @Override
        protected List<AnnotationMirror> defaultAction(Tree node,
                List<AnnotationMirror> p) {
            return p;
        }
        
        private Collection<? extends AnnotationMirror> annotateIfStatic(
                Element elt) {
            
            if (elt.getKind().isClass() || elt.getKind().isInterface()
                    || elt.getModifiers().contains(Modifier.STATIC))
                return Collections.singletonList(NONNULL);
            
            return Collections.emptyList();
        }
        
        @Override
        public List<AnnotationMirror> visitMemberSelect(MemberSelectTree node,
                List<AnnotationMirror> p) {
            
            Element elt = elementFromUse(node);
            assert elt != null;
            
            p.addAll(annotateIfStatic(elt));
            return super.visitMemberSelect(node, p);
        }


        @Override
        public List<AnnotationMirror> visitIdentifier(IdentifierTree node,
                List<AnnotationMirror> p) {
            
            Element elt = elementFromUse(node);
            assert elt != null;
            
            p.addAll(annotateIfStatic(elt));
            return super.visitIdentifier(node, p);
        }
        
        @Override
        public List<AnnotationMirror> visitNewArray(NewArrayTree node,
                List<AnnotationMirror> p) {
            p.add(NONNULL);
            return super.visitNewArray(node, p);
        }

        @Override
        public List<AnnotationMirror> visitNewClass(NewClassTree node,
                List<AnnotationMirror> p) {
            p.add(NONNULL);
            return super.visitNewClass(node, p);
        }

        @Override
        public List<AnnotationMirror> visitLiteral(LiteralTree node,
                List<AnnotationMirror> p) {
            if (node.getKind() == Tree.Kind.STRING_LITERAL)
                p.add(NONNULL);
            if (node.getKind() == Tree.Kind.NULL_LITERAL)
                p.add(NULLABLE);
            return super.visitLiteral(node, p);
        }
        
    }
    
    protected class ElementPreAnnotator extends SimpleElementVisitor6<List<AnnotationMirror>, List<AnnotationMirror>> {

        @Override
        protected List<AnnotationMirror> defaultAction(Element e,
                List<AnnotationMirror> p) {
            return p;
        }

        @Override
        public List<AnnotationMirror> visitPackage(PackageElement e,
                List<AnnotationMirror> p) {
            p.add(NONNULL);
            return super.visitPackage(e, p);
        }
        
    }
}
