package checkers.types;

import checkers.types.AnnotatedTypeMirror.*;
import checkers.basetype.BaseTypeChecker;

/**
 * A AnnotatedTypeScanner that compares a pair of annotations tree at each node.
 */
public class AnnotatedTypePairScanner extends AnnotatedTypeScanner<Boolean, AnnotatedTypeMirror> {

    protected BaseTypeChecker checker;

    /**
     * The constructor receives a BaseTypeChecker for performing checks
     * on subtypes, ignoring type parameters.
     */
    public AnnotatedTypePairScanner(BaseTypeChecker checker) {
        this.checker = checker;
    }

    @Override
    public final Boolean scan(AnnotatedTypeMirror type, AnnotatedTypeMirror p) {
        return checker.isSubtypeIgnoringTypeParameters(type, p);
    }

    @Override
    protected Boolean reduce(Boolean r1, Boolean r2) {
        if (r1 == null || r2 == null) return false;
        return r1 & r2;
    }

    @Override
    public Boolean visitDeclared(AnnotatedDeclaredType type, AnnotatedTypeMirror p) {
        if (! (p instanceof AnnotatedDeclaredType)) return false;
        boolean r = scan(type, p);
        if (!r) return false;
        AnnotatedDeclaredType type2 = (AnnotatedDeclaredType) p;
        for (int i = 0; i < type.getTypeArguments().size(); i++)
            r &= scan(type.getTypeArguments().get(i), type2.getTypeArguments().get(i));
        return r;
    }

    @Override
    public Boolean visitExecutable(AnnotatedExecutableType type, AnnotatedTypeMirror p) {
        if (! (p instanceof AnnotatedExecutableType)) return false;
        AnnotatedExecutableType type2 = (AnnotatedExecutableType) p;
        boolean r = scan(type.getReturnType(), type2.getReturnType());
        r &= scan(type.getReceiverType(), type2.getReceiverType());
        for (int i = 0; i < type.getParameterTypes().size(); i++)
            r &= scan(type.getParameterTypes().get(i),
                      type2.getParameterTypes().get(i));
        for (int i = 0; i < type.getThrownTypes().size(); i++)
            r &= scan(type.getThrownTypes().get(i),
                      type2.getThrownTypes().get(i));
        for (int i = 0; i < type.getTypeVariables().size(); i++)
            r &= scan(type.getTypeVariables().get(i),
                      type2.getTypeVariables().get(i));
        return r;
    }

    @Override
    public Boolean visitTypeVariable(AnnotatedTypeVariable type, AnnotatedTypeMirror p) {
        if (! (p instanceof AnnotatedTypeVariable)) return false;
        AnnotatedTypeVariable type2 = (AnnotatedTypeVariable) p;
        boolean r = scan(type.getLowerBound(), type2.getLowerBound());
        r &= scan(type.getUpperBound(), type2.getUpperBound());
        return r;
    }

    @Override
    public Boolean visitNoType(AnnotatedNoType type, AnnotatedTypeMirror p) {
        return true;
    }

    @Override
    public Boolean visitNull(AnnotatedNullType type, AnnotatedTypeMirror p) {
        return true;
    }

    @Override
    public Boolean visitPrimitive(AnnotatedPrimitiveType type, AnnotatedTypeMirror p) {
        return true;
    }

    @Override
    public Boolean visitWildcard(AnnotatedWildcardType type, AnnotatedTypeMirror p) {
        if (p instanceof AnnotatedWildcardType) {
            AnnotatedWildcardType type2 = (AnnotatedWildcardType) p;
            boolean r = scan(type.getExtendsBound(), type2.getExtendsBound());
            r &= scan(type.getSuperBound(), type2.getSuperBound());
            return r;
        } else {
            boolean r = true;

            AnnotatedTypeMirror ext = type.getExtendsBound();
            if (ext != null)
                r &= scan(p, ext);

            AnnotatedTypeMirror sup = type.getSuperBound();
            if (sup != null)
                r &= scan(sup, p);

            return r;
        }
    }
}
